#include <Xm/Xm.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <errno.h>

#include "xmps.h"


typedef struct _BufferRec {
     String data;
     int len;
} BufferStruct, *BufferRec;


typedef struct _CommandRec {
     XtInputId outputId;
     int outputPipe[2];
     BufferStruct outputBuffer;
     XtInputId errorId;
     int errorPipe[2];
     BufferStruct errorBuffer;
} CommandStruct, *CommandRec;


Boolean childDied (clientData)
     XtPointer clientData;
{
     int *pid = (int *) clientData;
     
     /*
      * This function just prevents XtProcessEvent() to hang if
      * our child dies when it is being executed and there is no
      * data left in the pipes : XtProcessEvent() will exit after
      * calling that function, instead of blocking on select().
      */

     return (waitpid (*pid, NULL, WNOHANG));
}

	  
void readInput (pipe, buffer)
     int pipe;
     BufferRec buffer;
{
     char tempbuf[BUFSIZ];
     long charsCount;
     int oldPipeMode;

     /*
      * Set the read mode of the pipe to non-blocking.
      */
     if ((oldPipeMode = fcntl (pipe, F_GETFL)) == -1) {
	  error ("Cannot get ouput of ps", (char *) 0);
	  return;
     }
     if (fcntl (pipe, F_SETFL, (oldPipeMode | O_NONBLOCK)) == -1) {
	  error ("Cannot get output of ps", (char *) 0);
	  return;
     }
     
     /*
      * Read as many characters as are available.
      */
     charsCount = read (pipe, tempbuf, BUFSIZ);
     while ((charsCount != 0) && (charsCount != -1)) {
	  buffer->data = XtRealloc (buffer->data,
				    (Cardinal) (buffer->len + charsCount + 1));
	  bcopy (tempbuf, buffer->data + buffer->len, charsCount);
	  buffer->len = buffer->len + charsCount;

	  charsCount = read (pipe, tempbuf, BUFSIZ);
     }

     /*
      * Restore the old pipe's mode, and return the output of ps.
      */
     (void) fcntl (pipe, F_SETFL, oldPipeMode);

     if ((charsCount == -1) && (errno != EAGAIN)) {
	  error ("Cannot read output of ps", (char *) 0);
	  return;
     }

     if (buffer->data != NULL)
	  buffer->data[buffer->len] = '\0';
     return;
}


XtInputCallbackProc inputReady (clientData, pipe, unused)
     XtPointer clientData;
     int *pipe;
     XtInputId *unused;
{
     BufferRec buffer = (BufferRec) clientData;

     readInput (*pipe, buffer);
}


String executeCommand (argv)
     String *argv;
{
     CommandStruct command;
     int oldStdout, oldStderr;
     int pid;
     
     command.outputBuffer.data = NULL;
     command.outputBuffer.len = 0;
     command.errorBuffer.data = NULL;
     command.errorBuffer.len = 0;

     /*
      * Redirect stdout to a pipe.
      */
     if (pipe (command.outputPipe) != -1) {
	  oldStdout = dup (fileno (stdout));
	  (void) dup2 (command.outputPipe[1], fileno (stdout));
     } else {
	  error ("Cannot redirect standard output", (char *) 0);
	  return (XtNewString (" \n"));
     }
     
     /*
      * Redirect stderr to a pipe too.
      */
     if (pipe (command.errorPipe) != -1) {
	  oldStderr = dup (fileno (stderr));
	  (void) dup2 (command.errorPipe[1], fileno (stderr));
     } else {
	  error ("Cannot redirect standard error", (char *) 0);
	  return (XtNewString (" \n"));
     }
     
     /*
      * Create a sub-process that will run ps.
      */
     if ((pid = fork ()) == -1) {
	  error ("Cannot fork", (char *) 0);
	  return (XtNewString (" \n"));

     } else if (pid == 0) {
	  /*
	   * Close unused file descriptors.
	   */
	  close (fileno (stdin));
	  close (oldStdout);
	  close (oldStderr);
	  close (command.outputPipe[0]);
	  close (command.errorPipe[0]);

	  /*
	   * Execute ps.
	   */
	  (void) execv (argv[0], argv);

	  /*
	   * If we ever reach this point, then exec failed.
	   */
	  error ("Cannot run ps", (char *) 0);
	  return (XtNewString (" \n"));

     } else {
	  /*
	   * Restore the original stdout and stderr.
	   */
	  dup2 (oldStdout, fileno (stdout));
	  dup2 (oldStderr, fileno (stderr));

	  /*
	   * Close unused file descriptors.
	   */
	  close (oldStdout);
	  close (oldStderr);
	  close (command.outputPipe[1]);
	  close (command.errorPipe[1]);

	  /*
	   * Install handlers to take care of pipes overflow.
	   */
	  command.outputId = XtAddInput (command.outputPipe[0],
					 (XtPointer) XtInputReadMask,
					 (XtInputCallbackProc) inputReady,
					 (XtPointer) &command.outputBuffer);
	  command.errorId = XtAddInput (command.errorPipe[0],
					(XtPointer) XtInputReadMask,
					(XtInputCallbackProc) inputReady,
					(XtPointer) &command.errorBuffer);

	  /*
	   * Install a handlers to take care of dead locks.
	   */
	  XtAddWorkProc ((XtWorkProc) childDied, &pid);

	  /*
	   * Read as much as we can to prevent pipe overflows
	   * whilst our child is still alive.
	   */
	  while (!waitpid (pid, NULL, WNOHANG))
	       XtProcessEvent (XtIMAlternateInput);

	  (void) waitpid (pid, NULL, 0);
	  
	  /*
	   * Get the output of ps, if any.
	   */
	  readInput (command.outputPipe[0], &command.outputBuffer);

	  (void) close (command.outputPipe[0]);
	  XtRemoveInput (command.outputId);

	  /*
	   * Check for any errors.
	   */
	  readInput (command.errorPipe[0], &command.errorBuffer);

	  (void) close (command.errorPipe[0]);
	  XtRemoveInput (command.errorId);

	  if (command.errorBuffer.data != NULL) {
	       error (command.errorBuffer.data, (char *) 0);
	       XtFree (command.errorBuffer.data);
	  }

	  /*
	   * Return the output of ps.
	   */
	  return (command.outputBuffer.data);
     }
}
